# Built-in
import os
import logging
import subprocess
import sys
import time
import socket

# Internal
from nxt import nxt_log
from . import nxt_io
from . import nxt_path
from . import nxt_layer
from nxt.remote import get_running_server_address
from nxt.remote.client import NxtClient
from .remote import contexts
from nxt.stage import Stage
from nxt.constants import NXT_DCC_ENV_VAR, is_standalone

logger = logging.getLogger(__name__)


class Session(object):
    """Nxt Session, manages loading, saving, and execution of graphs."""
    def __init__(self):
        self._loaded_files = {}
        self.log_file = nxt_log.make_session_log()
        self.rpc_server = None

    @property
    def loaded_files(self):
        return self._loaded_files

    def new_file(self):
        """Create a new (empty) graph.

        :return: New graph object.
        """
        new_graph = Stage(name=self.get_unused_graph_name('untitled'))
        self._loaded_files[new_graph.uid] = new_graph
        return new_graph

    def get_unused_graph_name(self, name):
        """Used by new_file() to make sure that any new graph created won't
        conflict with the names of currently open graphs.

        :param name: Name preference, adjusted to remove conflicts if needed.
        :type name: str

        :return: Resulting name, may not match input name preference.
        """
        def test_name(n):
            for stage in self._loaded_files.values():
                if n == stage._name:
                    return False
            return True

        result_name = ''
        num_suffix = 0
        potential_name = name
        while result_name == '':
            if test_name(potential_name):
                result_name = potential_name
            else:
                num_suffix += 1
                potential_name = name + str(num_suffix)

        return result_name

    def load_file(self, filepath):
        """Create a new graph from given filepath.

        :param filepath: Path to .json file to load.
        :type filepath: str

        :return: New graph object, if load succeeded. Otherwise None.
        :rtype: Graph
        """
        try:
            layer_data = nxt_io.load_file_data(filepath)
            new_stage = Stage(layer_data=layer_data)
        except IOError:
            logger.exception('Failed to open: "{}"'.format(filepath))
            raise
        self._loaded_files[new_stage.uid] = new_stage
        return new_stage

    def unload_file(self, uid):
        unload_key = [k for k in self.loaded_files.keys() if k == uid]
        uid_lst = [k for k in self.loaded_files.keys()]
        if not unload_key:
            err_format = ('Unable to find a graph with the uuid {uid} '
                          'exsisting uuids are {uidLst}')
            raise LookupError(err_format.format(uid=uid, uidLst=uid_lst))
        unload_key = unload_key[0]
        self.loaded_files.pop(unload_key)

    def save_file(self, graph_file_path, new_path=None):
        """The file to save is specified via graph_file_path.
        If new_path is specified, it will change the path used to
        identify the graph in the future.

        :param graph_file_path: path used to locate file to save in currently
                                loaded files.
        :type graph_file_path: str

        :param new_path: If specified, saves the specified file to a new path.
        :type new_path: str

        :return: True/False if successful
        """
        target_path = graph_file_path
        if new_path:
            target_path = os.path.normcase(new_path)
        current_path = None
        if graph_file_path:
            current_path = os.path.normcase(graph_file_path)
        if current_path in (k[0] for k in self._loaded_files.keys()):
            self._loaded_files[current_path].save_layer(filepath=target_path)
            if current_path != target_path:
                changed_graph = self._loaded_files.pop(current_path)
                self._loaded_files[target_path] = changed_graph
            return True
        else:
            return False

    def save_layer(self, layer, filepath=None):
        layer.save(filepath=filepath)

    def get_stage(self, path):
        """Get the graph object for a specified file path.
        If the file is not already loaded, it will be.

        :param path: Path to file the graph is loaded from.
        :type path: str

        :return: Graph object, if found, otherwise None.
        """
        norm_path = os.path.normcase(path)
        for uid, data in self._loaded_files.items():
            if data.filepath == norm_path:
                return self._loaded_files[uid]
        else:
            return self.load_file(path)

    def execute_graph(self, filepath, start=None, parameters=None,
                      context=None):
        """Execute the graph at the given file path. Optionally at given
        start. You may provided parameters to the parameter arg. The data
        should be formatted as follows::
            {
                '/node.attr': 'New value!',
                '/another.count: 5
            }
        If the node does not exist an error will be logged but the execution
        will continue. If the attr does not exist (but the node does) it
        will be created.

        :param filepath: Path to the graph file.
        :type filepath: str
        :param start: Path to the node to begin execution from OR int \
        of start node index
        :type start: str or int
        :param parameters: Dict where key is attr path and value is new attr \
        value.
        :type parameters: dict
        :param context: Optional name of remote context to execute graph in, \
        if none is passed the graph is executed in this interpreter.
        :type context: str
        :return: a runtime CompLayer.
        """
        stage = self.get_stage(filepath)
        if context is None:
            self.start_rpc_if_needed(stage)
            try:
                return stage.execute(start=start, parameters=parameters)
            finally:
                self.shutdown_rpc_server()
        else:
            self._start_rpc_server()
            proxy = NxtClient()
            try:
                cache_file = proxy.exec_in_headless(filepath, start, None,
                                                    parameters, context)
                return nxt_layer.CacheLayer.load_from_filepath(cache_file)
            finally:
                self.shutdown_rpc_server()

    def execute_nodes(self, filepath, node_paths, parameters=None):
        stage = self.get_stage(filepath)
        self.start_rpc_if_needed(stage)
        comp_layer = stage.build_stage()
        try:
            return stage.execute_nodes(node_paths=node_paths,
                                       layer=comp_layer, parameters=parameters)
        finally:
            self.shutdown_rpc_server()

    def start_rpc_if_needed(self, stage):
        """Given a stage this method parses for the remote context node,
        if it is found in the stage's sub_layers an rpc is started.
        (if one is already run a new on will NOT be started)

        :param stage: stage.Stage instance
        :return: None
        """
        # Remote Graphs
        remote_node_name = contexts.REMOTE_CONTEXT_BUILTIN_NODE
        remote_path = nxt_path.join_node_paths(nxt_path.NODE_SEP,
                                               remote_node_name)
        # Sub-graphs that might call remote graphs
        sub_graph_node_name = contexts.SUB_GRAPH_BUILTIN_NODE
        sub_graph_path = nxt_path.join_node_paths(nxt_path.NODE_SEP,
                                                  sub_graph_node_name)
        start_rpc = False
        for layer in stage._sub_layers:
            if layer.lookup(remote_path) or layer.lookup(sub_graph_path):
                start_rpc = True
                break
        if start_rpc:
            self._start_rpc_server()
            return
        logger.info('It was determined you do not need an rpc server.')

    def _start_rpc_server(self, custom_stdout=False, rpc_log_filepath=None,
                          socket_log=False, stream_handler=None):
        """Directly start the RPC server regardless of need for it.
        :param custom_stdout: If True the stdout and stderr of the rpc server
        will be directed to a file on disc
        :param rpc_log_filepath: Optional filepath to be used for custom_stdout.
        The file must be writeable. If custom_stdout is False this kwarg has
        no effect.
        :param socket_log: If True the a nxt logging messages will be
        directed to a socket handler.
        :return: None
        """
        logger.debug('Starting rpc server!')
        rpc_server = RPCServerProcess.start(use_custom_stdout=custom_stdout,
                                            stdout_filepath=rpc_log_filepath,
                                            socket_log=socket_log,
                                            stream_handler=stream_handler)
        self.rpc_server = rpc_server

    def shutdown_rpc_server(self):
        """Attempts to close THIS session's rpc server, if this session
        doesn't own the server it will not shut it down.
        :return: None
        """
        if not self.rpc_server:
            logger.debug('This session does not own an rpc server.')
            return
        logger.debug('Trying to shutdown rpc server...')
        try:
            self.rpc_server.terminate()
        except:
            logging.exception('Failed to shut down rpc server!')
            pass
        self.rpc_server = None


class RPCServerProcess(object):
    STARTUP_TIMEOUT = 5

    def __init__(self, use_custom_stdout=False, stdout_filepath=None,
                 socket_log=False, stream_handler=None):
        """Uses a subprocess to start remote/server.py
        If it is determined there is already a server running somewhere on
        the host address then run will early exit. If for some reason the
        port is occupied by another process that is not our RPC server an
        exception is raised.
        :param use_custom_stdout: If True the stdout/stderr of the subprocess
        is populated by a file handle to a custom filepath.
        :param stdout_filepath: Optionally, if you don't want a generated
        temp file as the stdout file a path to your desired output file can
        be passed here. Has no effect if use_custom_stdout is False.
        :param socket_log: If True the a nxt logging messages will be
        :param stream_handler: Logging stream handler only needed if calling
        from visual app.
        directed to a socket handler.
        """
        self.terminal = None
        self.use_custom_stdout = use_custom_stdout
        self.server_log_file = stdout_filepath or ''
        self.socket_logging = socket_log
        self.stream_handler = stream_handler

    @classmethod
    def start(cls, use_custom_stdout=False, stdout_filepath=None,
              socket_log=False, stream_handler=None):
        """Quick start method for starting and returning a server if needed

        :param use_custom_stdout: see __init__
        :param stdout_filepath: see __init__
        :param socket_log: see __init__
        :return: None or RPCServerProcess instance
        """
        if not is_standalone():
            logger.warning('The nxt rpc server cannot be started unless nxt '
                           'is launched as standalone.')
            return
        rpc_server = cls(use_custom_stdout=use_custom_stdout,
                         stdout_filepath=stdout_filepath,
                         socket_log=socket_log, stream_handler=stream_handler)
        if rpc_server.run():
            return rpc_server

    def run(self):
        """Method for starting rpc server sub-process

        :raises: OSError
        :return: True if server was started, else False.
        """
        if self.is_running():
            logger.info('Server already running somewhere...')
            return False
        elif not self.is_port_available():
            _, port = get_running_server_address(as_str=False)
            raise OSError('Port {} is not available!'.format(port))
        old_env_verbosity = os.environ.get(nxt_log.VERBOSE_ENV_VAR, None)
        if self.socket_logging:
            os.environ[nxt_log.VERBOSE_ENV_VAR] = 'socket'
        if self.use_custom_stdout:
            if not self.server_log_file:
                _ext = '.nxtlog'
                self.server_log_file = nxt_io.generate_temp_file(suffix=_ext)
                nxt_log.track_log_file(self.server_log_file)
            server_log_handle = open(self.server_log_file, 'w')
            nxt_log.startup_log_socket(stream_handler=self.stream_handler)
        else:
            server_log_handle = None
        call = [sys.executable, '-m', 'nxt.remote.server', self.server_log_file]
        logger.debug("Calling: ")
        logger.debug(str(call))
        self.terminal = subprocess.Popen(call, stdout=server_log_handle,
                                         stderr=server_log_handle)
        count = 0
        while count < self.STARTUP_TIMEOUT:
            logger.debug('Waiting on rpc server...')
            time.sleep(1)
            count += 1
            if self.is_running():
                break
        if count == self.STARTUP_TIMEOUT:
            raise OSError('Failed to start RPC server!')
        logger.debug('rpc server started')
        if old_env_verbosity is not None:
            os.environ[nxt_log.VERBOSE_ENV_VAR] = old_env_verbosity
        return True

    def is_running(self):
        """Attempts to connect to the rpc server as a client, if that fails
        because the connection is refused the subprocess is polled. If this
        instance doesn't own the subprocess and it fails to connect as a
        client False is returned.

        :return: bool of server running state
        """
        is_running = False
        try:
            proxy = NxtClient()
            is_running = proxy.is_alive()
        except Exception as e:
            # Connection refused
            if getattr(e, 'errno', -1) == 10061 and self.terminal:
                poll = self.terminal.poll()
                if poll is None:
                    is_running = True
        return is_running

    def terminate(self):
        """Attempts to kill the rpc server.
        Note: The server is killed even if it isn't owned by this class!

        :return: None
        """
        try:
            proxy = NxtClient()
            proxy.kill()
        except Exception as e:
            if getattr(e, 'errno', -1) == 10061 and self.terminal:
                try:
                    logger.info('Telling rpc server to shutdown...')
                    proxy = NxtClient()
                    proxy.kill()
                    self.server_log_file = ''
                except:
                    logger.warning(
                        'Unable to tell rpc server to shutdown!')
                try:
                    self.terminal.terminate()
                    self.terminal = None
                    logger.debug('RPC server has shutdown!')
                except:
                    logger.warning('Unable to kill rpc process!')
        nxt_log.shutdown_log_socket()

    @staticmethod
    def is_port_available():
        """Checks if the rpc port is occupied.
        :return: bool
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        address = get_running_server_address(as_str=False)
        results = s.connect_ex(address)
        s.close()
        if results == 0:
            return False
        return True



# Builtin
import sys
import os
import logging
from collections import namedtuple

from nxt import nxt_io

logger = logging.getLogger('nxt')

REMOTE_CONTEXT_BUILTIN_NODE = '_remote_sub_graph'
SUB_GRAPH_BUILTIN_NODE = '_sub_graph'
REMOTE_CONTEXT_ATTR_NAME = '_context'


class RemoteContext(object):
    def __init__(self, name, exe, graph, args=()):
        """Create a cusom remote context for NXT, the return of this function
        should be fed to `register_context`.

        :param name: Desired context name (used to identify it to users)
        :param exe: Path to Python interpreter executable, if another executable
        is supplied a script to call with that executable is expected.
        :param graph: Filepath of context setup graph
        :param args: If the exe arg is not a Python interpreter additional args
        may need to passed.
        :type args: list | tuple
        """
        self.name = name
        self.exe = exe
        self.graph = graph
        self.args = args


_CONTEXTS = []
starter_contexts = {'maya':
                        os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                     'maya_context.nxt')),
                    'blender':
                        os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                     'blender_context.nxt'))}


def register_context(context):
    """Register context class with nxt

    >>> my_context = RemoteContext('Maya', 'path_to/bin/mayapy.exe', './maya_context.nxt')
    (RemoteContext)
    >>> register_context(my_context)
    (registered context: Maya)

    :param context: RemoteContext object
    :type context: RemoteContext
    """
    global _CONTEXTS
    _CONTEXTS += [context]
    logger.info('registered context: ' + context.name)


def find_context_by_name(name):
    for context in _CONTEXTS:
        if context.name == name:
            return context
    return None


def iter_context_names():
    for context in _CONTEXTS:
        yield context.name


_context_graph = '${var}/_context.nxt'.format(var=nxt_io.BUILTIN_GRAPHS_ENV_VAR)
PYTHON_CONTEXT = RemoteContext('python', sys.executable, _context_graph)
register_context(PYTHON_CONTEXT)


def get_current_context_exe_name():
    exe_name = os.path.basename(sys.executable)
    exe_name, _ = os.path.splitext(exe_name)
    return exe_name



"""Utitiles for parsing and create nxt token strings.
"""
# Built-in
import logging
import sys
from collections import namedtuple

logger = logging.getLogger('nxt.tokens')

TOKEN_PREFIX = '${'
TOKEN_SUFFIX = '}'

Token = namedtuple('Token', ['prefix', 'detect', 'resolve'])

plugin_tokens = []


def register_token(prefix, detect, resolve):
    """Define a token to be expandable by resolve. Requires 3 components, a \
        prefix, a detect callable, and a resolve callable.

    For an example token see the example plugin in the nxt plugins directory.

    :param prefix: prefix of token
    :type prefix: str
    :param detect: callable that takes a single argument and returns True if \
        given input can be resolved by this token type.
    :type detect: callable
    :param resolve: callable that resolves token content
    :type resolve: callable
    """
    logger.info('Registered token: ' + prefix)
    new_token = Token(prefix, detect, resolve)
    global plugin_tokens
    plugin_tokens += [new_token]
    return new_token


class TOKENTYPE(object):
    # TODO: Make these fully defined tokens, right now we're just using them
    #  to unify how we get token prefixes.
    ATTR = Token('', None, None)
    FILE = Token('file::', None, None)
    FILEPATH = Token('path::', None, None)
    PATH = Token('_nxtpath', None, None)
    CONTENTS = Token('contents::', None, None)
    COLOR = Token('_nxtcolor', None, None)
    # ATTR is at the end because it does not have a prefix
    ALL = (FILE, FILEPATH, PATH, CONTENTS, COLOR, ATTR)


def get_token_content(token_str):
    """Removes token prefix and suffix from given token_str

    :param token_str: string to remove token syntax from
    :type token_str: str
    :return: given token_str with token syntax removed
    :rtype: str
    """
    _, _, no_prefix = token_str.partition(TOKEN_PREFIX)
    content, _, _ = no_prefix.rpartition(TOKEN_SUFFIX)
    return content


def make_token_str(token_content):
    """Build a token around given token_content.
    The opposite of `get_token_content`

    :param token_content: token content to wrap
    :type token_content: str
    :return: token content wrapped in token syntax
    :rtype: str
    """
    return '{}{}{}'.format(TOKEN_PREFIX, token_content, TOKEN_SUFFIX)


def get_standalone_tokens(raw_value, token_types=None):
    """Get tokens in the given value that are not nested within another token.
    If none are found, an empty list is returned.
    If the token syntax is malformed(extra starts or ends), return empty list.
    Returns list including outer token syntax

    :param raw_value: value to find standalone tokens in.
    :type raw_value: str
    :param token_types: Optionally a list of specific token types can be \
    provided, only standalone tokens that look like those token type(s) will \
    be returned.
    :type token_types: list, tuple, or None
    :return: list of standalone tokens, if found.
    :rtype: list
    """
    # Early exit if the raw value is None or not a string like object
    if sys.version_info[0] == 2:
        if not raw_value or not isinstance(raw_value, basestring):
            return []
    else:
        if not raw_value or not isinstance(raw_value, str):
            return []
    if token_types is None:
        token_types = TOKENTYPE.ALL
    i = 0
    bounds = []
    starts = 0
    ends = 0
    outer_start_idx = -1
    for char in raw_value:
        # First condition verifies we're not at the end of the string, where
        # it would be impossible to get second prefix character.
        if (i+1 < len(raw_value) and
           char == TOKEN_PREFIX[0] and raw_value[i+1] == TOKEN_PREFIX[1]):
            if starts == 0:
                outer_start_idx = i
            starts += 1
        elif starts and char == TOKEN_SUFFIX:
            ends += 1
        if starts > 0 and starts == ends:
            bounds += [(outer_start_idx, i+1)]
            starts = 0
            ends = 0
            outer_start_idx = -1
        i += 1
    if starts != ends:
        return []
    results = []
    for start, end in bounds:
        skip = True
        token_contents = raw_value[start:end]
        for token_type in token_types:
            prefix = token_type.prefix
            if token_contents.startswith(TOKEN_PREFIX+prefix):
                skip = False
        if skip:
            continue
        results += [token_contents]
    return results


def atomic_token_partition(value):
    """Partition given value on a token that appears resolvable(contains no
    sub tokens). Returns in a tuple: (before_token, token, after_token).
    Returned token includes token syntax. If no tokens are found, returned
    tuple contains None in all values.

    :param value: text to find a token from, and partition
    :type value: str
    :return: before_token, token, after_token
    :rtype: tuple(str, str, str)
    """
    before, sep, after_bef = value.rpartition(TOKEN_PREFIX)
    if not sep:
        return (None, None, None)
    token, sep, after = after_bef.partition(TOKEN_SUFFIX)
    if not sep:
        # msg = 'bad resolve formatting, cannot find closer for {}'
        # msg = msg.format(before + tokens.TOKEN_PREFIX)
        # logger.error(msg)
        return (None, None, None)
    return before, make_token_str(token), after


def get_atomic_tokens(value):
    """Return a list of all tokens that appear ready to resolve.

    :param value: string to find tokens in
    :type value: str
    :return: tokens found
    :rtype: list
    """
    found = []
    prefix, content, suffix = atomic_token_partition(value)
    if not content:
        return found
    found += [content]
    if prefix:
        found += get_atomic_tokens(prefix)
    if suffix:
        found += get_atomic_tokens(suffix)
    found = set(found)
    return found





"""Common nxt pathing operations. Both for file and node paths.
"""
# Built-in
import os
import re
import sys
import logging

logger = logging.getLogger(__name__)


def full_file_expand(path, start=None):
    """A combination of commonly used os.path functions called to completely
    expand given `path`. If `start` is given, os.cwd is temporarily changed
    in order to influence the start of relative path expansion.
    Return value will always use forward slashes as seperator.

    :param path: path to expand
    :type path: str
    :param start: base directory to expand from, defaults to os cwd
    :type start: str, optional
    :return: given path expanded. Relative paths, environment \
    variables, and user character "~" are expanded.
    :rtype: [type]
    """
    orig_cwd = os.getcwd()
    if start:
        os.chdir(start)

    full_path = unify_env_vars(path)
    full_path = os.path.expandvars(full_path)
    full_path = os.path.expanduser(full_path)
    full_path = os.path.realpath(full_path)
    os.chdir(orig_cwd)
    full_path = full_path.replace(os.path.sep, '/')  # Thanks windows.
    return full_path


# Regex patterns for finding all env vars.
# NOTE match uses platform specific identifier, returns name only in group1.
PLATFORM_VAR_PATTERN_MAP = {
    'linux': re.compile(r"\$([a-zA-Z_]*)"),
    'win32': re.compile(r"%([a-zA-Z_]*)%")
}
PLATFORM_VAR_PATTERN_MAP['linux2'] = PLATFORM_VAR_PATTERN_MAP['linux']
PLATFORM_VAR_PATTERN_MAP['darwin'] = PLATFORM_VAR_PATTERN_MAP['linux']

# Strings ready to format with env var name. Result is ready for os expansion.
PLATFORM_VAR_FORMAT_MAP = {
    'linux': "${}",
    'win32': "%{}%"
}
PLATFORM_VAR_FORMAT_MAP['linux2'] = PLATFORM_VAR_FORMAT_MAP['linux']
PLATFORM_VAR_FORMAT_MAP['darwin'] = PLATFORM_VAR_FORMAT_MAP['linux']

# Strings ready to format with env var name into regex pattern for var.
PLATFORM_FIND_VAR_FORMAT_MAP = {
    'linux': r"\${}(?![a-zA-Z_])",
    'win32': r"%{}%"
}
PLATFORM_FIND_VAR_FORMAT_MAP['linux2'] = PLATFORM_FIND_VAR_FORMAT_MAP['linux']
PLATFORM_FIND_VAR_FORMAT_MAP['darwin'] = PLATFORM_FIND_VAR_FORMAT_MAP['linux']


def unify_env_vars(path, target_platform=sys.platform):
    """Replace foreign platform environment variables with platform native ones
    Environment variable patterns must be defined in PLATFORM_VAR_PATTERN_MAP
    and PLATFORM_VAR_PATTERN_MAP to be able to find and replace.

    In short, if you're on windows, convert linux style environment vars to
    windows style. If you're on linux, the reverse.

    Supports mixed type environment variables.

    :param path: string to replace variables in
    :type value: str
    :param target_platform: platform to unify as, defaults to sys.platform
    :type target_platform: str, optional
    :return: given path as unified as possible
    :rtype: str
    """
    target_search_pattern = PLATFORM_VAR_PATTERN_MAP[target_platform]
    replaced_patterns = []  # patterns we've alreay replaced.
    # replace platform vars from other platforms with vars from target platform
    result = path
    for platform, pattern in PLATFORM_VAR_PATTERN_MAP.items():
        # Do not replace vars from our target platform
        if pattern == target_search_pattern:
            continue
        # Do not replace a pattern more than once.
        if pattern in replaced_patterns:
            continue
        replaced_patterns += [pattern]
        found_vars = re.findall(pattern, path)
        for var in found_vars:
            bad_var = PLATFORM_VAR_FORMAT_MAP[platform].format(var)
            new_var = PLATFORM_VAR_FORMAT_MAP[target_platform].format(var)
            result = result.replace(bad_var, new_var)
    return result


# NODE SECTION
NODE_SEP = '/'
ATTR_SEP = '.'
WORLD = NODE_SEP


def expand_relative_node_path(rel_path, start_node_path):
    """Expands given node `rel_path` using `start_node_path` as root of
    relative path. This function does not validate the path, it is only a
    string machine.

    :param rel_path: relative path to expand.
    :type rel_path: str
    :param start_node_path: node to treat as root of relative path.
    :type start_node_path: str
    :return: expanded node path
    :rtype: str
    """
    if not rel_path:
        return None
    if rel_path == WORLD:
        return rel_path
    current_path = start_node_path
    if rel_path.startswith(NODE_SEP):
        return rel_path
    split_path = rel_path.split(NODE_SEP)
    for directive in split_path:
        if directive == '..':
            current_path, _, _ = current_path.rpartition(NODE_SEP)
            continue
        if directive == '.':
            continue
        if directive == '':
            continue
        # If we get this far, it must be the name of something.
        current_path += NODE_SEP
        current_path += directive
    return current_path


def path_attr_partition(str_path):
    """Returns a tuple of the node path and attribute name from the given path.

    >>> nxt_path.node_path_partition('somewhere/else.attr')
    ('somewhere/else', 'attr')
    >>> nxt_path.node_path_partition('...attr')
    ('..', 'attr')
    >>> nxt_path.node_path_partition('..attr')
    ('.', 'attr')

    :param str_path: node or attribute path
    :type str_path: str
    :return: tuple of node path and attr name ('node/path', 'attr_name')
    :rtype: tuple
    """
    if not is_attr_path(str_path):
        # This is a node path.
        return str_path, None
    if str_path[-1] is ATTR_SEP:
        # This is a node path.
        return str_path, None
    node_path, _, attr_name = str_path.rpartition(ATTR_SEP)
    return node_path, attr_name


def str_path_to_node_namespace(str_path):
    """Given an absolute node or attr `str_path`, split it into the
    list-"namespace" format used by stage.
    NOTE that if given an attribute path, that attribute name will not be
    present in the returned path.
    """
    node_path, _ = path_attr_partition(str_path)
    if node_path.startswith(NODE_SEP):
        return node_path.split(NODE_SEP)[1:]
    return node_path.split(NODE_SEP)


def node_namespace_to_str_path(namespace):
    """Given a node namespace, return it rejoined into a full string path.
    """
    return NODE_SEP + NODE_SEP.join(namespace)


def node_path_from_attr_path(attr_path):
    """Given an attribute path, return the node path contained within.

    :param attr_path: path to attribute
    :type attr_path: str
    :return: node path
    :rtype: str
    """
    node_path, _ = path_attr_partition(attr_path)
    return node_path


def attr_name_from_attr_path(attr_path):
    """Given an attribute path, return the attribute name.

    :param attr_path: path to attribute
    :type attr_path: str
    :return: attr name contained within input
    :rtype: str
    """
    _, attr_name = path_attr_partition(attr_path)
    return attr_name


def node_name_from_node_path(node_path):
    """Given a node path, return the node name of specified node.

    :param node_path: str
    :type node_path: node path to get node name from
    :return: node name
    :rtype: str
    """
    if node_path == WORLD:
        return WORLD
    _, _, node_name = node_path.rpartition(NODE_SEP)
    return node_name


def get_parent_path(node_path):
    """Returns the parent path implied by given `node_path`

    :param node_path: str
    :type node_path: node path to get parent path from
    :return: parent path
    :rtype: str
    """
    if node_path == WORLD:
        return ''
    parent_path, _, _ = node_path.rpartition(NODE_SEP)
    if not parent_path:
        parent_path = NODE_SEP
    return parent_path


def get_root_path(node_path):
    """Returns the implied "root" node of the given path.
    A root is a node with no parent.

    :param node_path: str
    :type node_path: node path to get root path from
    :return: root path
    :rtype: str
    """
    root_name, _, _ = node_path[1:].partition(NODE_SEP)
    return NODE_SEP + root_name


def is_attr_path(path):
    """Whether the given `path` appears to be a path to an attribute.

    :param node_path: str
    :type node_path: path to check
    :return: True/False whether path appears to be an attribute path.
    :rtype: bool
    """
    return ATTR_SEP in path


def make_attr_path(node_path, attr_name):
    """Given a node path and attr name, join them into an attribute path.

    :param node_path: node path to join
    :type node_path: str
    :param attr_name: attr name to join
    :type attr_name: str
    :return: constructed attribute path
    :rtype: str
    """
    return node_path + ATTR_SEP + attr_name


def join_node_paths(node_path1, node_path2):
    """Given 2 node paths, treat node_path1 as ancestor, and combine their
    paths together

    :param node_path1: intended ancestor of given paths
    :type node_path1: str
    :param node_path2: intended descendant of given paths
    :type node_path2: str
    :return: path to descendant of node_path1
    :rtype: str
    """
    if not isinstance(node_path2, (tuple, list)):
        node_path2 = (node_path2,)
    extras = (node_path1,) + node_path2
    joined = NODE_SEP.join(extras).replace(NODE_SEP+NODE_SEP, NODE_SEP)
    return joined


def replace_ancestor(path, old_ancestor, new_ancestor):
    """Given a node path, replace given old ancestor path with new ancestor \
        path.

    :param path: Original path which will have its ancestor replaced
    :type path: str
    :param old_ancestor: Path to ancestor of `path` that will be replaced
    :type old_ancestor: str
    :param new_ancestor: New ancestor to replace `old_ancestor`
    :type new_ancestor: str
    :return: New path with `old_ancestor` replaced with `new_ancestor`
    :rtype: str
    """
    path = _add_path_terminator(path)
    old_ancestor = _add_path_terminator(old_ancestor)
    new_ancestor = _add_path_terminator(new_ancestor)
    new_path = new_ancestor + path
    old_ancestor_is_world = old_ancestor == WORLD
    if not old_ancestor_is_world:
        new_path = path.replace(old_ancestor, new_ancestor, 1)
    new_path = new_path[:-1]
    return new_path.replace(NODE_SEP+NODE_SEP, NODE_SEP)


def is_ancestor(path_to_check, ancestor_path):
    """Returns if given ancestor_path is an ancestor path to path_to_check

    :param path_to_check: potential descendant path
    :type path_to_check: str
    :param ancestor_path: potential ancestor path
    :type ancestor_path: str
    :return: Whether ancestor_path is ancestor of path_to_check
    :rtype: bool
    """
    if ancestor_path is WORLD:
        return True
    return path_to_check.startswith(ancestor_path + NODE_SEP)


def all_ancestor_paths(path):
    """Returns the ancestor paths implied by given path

    :param path: node path
    :type path: str
    :return: list of implied ancestor paths
    :rtype: list
    """
    ancestors = []
    if path == WORLD:
        return ancestors
    parent_path = get_parent_path(path)
    while parent_path is not WORLD:
        ancestors += [parent_path]
        parent_path = get_parent_path(parent_path)
    return ancestors


def _add_path_terminator(path):
    path += NODE_SEP
    return path


def get_path_depth(path):
    """Get depth of given path. Depth is number of nodes deep from world.
    World is depth 0.

    :param path: path to check depth of.
    :type path: str
    :return: depth of given path.
    :rtype: int
    """
    if path == WORLD:
        return 0
    return path.count(NODE_SEP)


def trim_to_depth(path, trim_depth):
    """Trim a path to given depth. Removes nodes from the end of the path to
    trim to an ancestor path of given path. If trim depth is 0, world path
    is returned. If trim depth is greater than or equal to the depth of given
    path, path is returned unchanged.

    :param path: path to trim
    :type path: str
    :param trim_depth: depth to trim to
    :type trim_depth: int
    :return: path, trimmed to trim depth if possible
    :rtype: str
    """
    if trim_depth == 0:
        return WORLD
    given_path_depth = get_path_depth(path)
    if given_path_depth <= trim_depth:
        return path
    remove_count = given_path_depth - trim_depth
    return path.rsplit(NODE_SEP, remove_count)[0]



"""Ways to operate on and ask questions of nxt layers.
"""
# Built-in
import os
import json
import copy
import logging
import time
import sys
from collections import OrderedDict

# Internal
if sys.version_info[0] == 2:
    import nxt_io
    import nxt_path
    import nxt_node
    from constants import GRAPH_VERSION
    from runtime import Console
else:
    from . import nxt_io
    from . import nxt_path
    from . import nxt_node
    from .constants import GRAPH_VERSION
    from .runtime import Console
from . import UNTITLED

logger = logging.getLogger(__name__)


class SAVE_KEY(object):
    VERSION = 'version'
    ALIAS = 'alias'
    MUTE = 'mute'
    SOLO = 'solo'
    REFERENCES = 'references'
    COMP_ORVERRIDES = 'comp_overrides'
    COLOR = 'color'
    META_DATA = 'meta_data'
    NODES = 'nodes'
    CWD = 'cwd'
    FILEPATH = 'filepath'
    REAL_PATH = 'real_path'
    ATTRS = 'attrs'
    NAME = 'name'
    LOCK = 'lock'


class META_DATA_KEY(object):
    POSITIONS = 'positions'
    COLLAPSE = 'collapse'
    COLORS = 'colors'
    ALIASES = 'aliases'
    LOCKS = 'locks'


class LAYERS(object):
    COMP = '<display>'
    TARGET = '<target>'
    TOP = '<top>'


class AUTHORING:
    CREATE = 'create'
    REFERENCE = 'reference'
    ABOVE = 0
    BELOW = 1


class LayerReturnTypes(object):
    Node = 'Node'  # Return a node or list of nodes
    Path = 'Path'  # Return a node path or list of node paths
    NodeTable = 'NodeTable'  # Return a list of lists [[NodePath, Node]]
    NameDict = 'NameDict'  # Return a dict where k is node NAME and v is node
    PathDict = 'PathDict'  # Return a dict where k is node PATH and v is node
    Boolean = 'Boolean'  # Return if there is at least one child


class SpecLayer(object):
    """Layer object representing layer data from disk.
    """
    RETURNS = LayerReturnTypes

    @classmethod
    def load_from_filepath(cls, filepath):
        """Given a filepath, return a SpecLayer loaded from that path.

        :param filepath: path to file to load
        :type filepath: str
        :return: Loaded SpecLayer
        :rtype: SpecLayer
        """
        layer_data = nxt_io.load_file_data(filepath)
        return cls.load_from_layer_data(layer_data)

    @classmethod
    def load_from_layer_data(cls, layer_data):
        return cls(layer_data)

    def __init__(self, layer_data=None):
        """Create a SpecLayer, optionally from layer data.

        :param layer_data: layer data dictionary, defaults to None
        :type layer_data: dict, optional
        """
        if layer_data is None:
            layer_data = {}
        self._node_table = []
        self._cached_children = {}
        self._cached_implied_children = {}
        self._name = UNTITLED
        self._layer_idx = 0
        self.mute = False
        self.solo = False
        self.filepath = layer_data.get(SAVE_KEY.FILEPATH)
        self.real_path = layer_data.get(SAVE_KEY.REAL_PATH)
        self.cwd = layer_data.get(SAVE_KEY.CWD)
        if self.filepath:
            file_name = os.path.basename(self.filepath)
        else:
            file_name = UNTITLED
        self.alias = layer_data.get(SAVE_KEY.ALIAS, file_name)
        self.comp_overrides = layer_data.get(SAVE_KEY.COMP_ORVERRIDES, {})
        self.sub_layer_paths = []
        self.sub_layers = []
        for layer_path in layer_data.get(SAVE_KEY.REFERENCES, []):
            if not layer_path:
                continue
            self.sub_layer_paths += [layer_path]
            self.sub_layers += [{SAVE_KEY.FILEPATH: layer_path}]
        self.parent_layer = layer_data.get('parent_layer', None)
        meta_data = layer_data.get(SAVE_KEY.META_DATA, {})
        # Fixme: Shouldn't this use the same method as solo/mute/lock?
        self.positions = meta_data.get(META_DATA_KEY.POSITIONS, {})
        self.collapse = meta_data.get(META_DATA_KEY.COLLAPSE, {})
        self.aliases = meta_data.get(META_DATA_KEY.ALIASES, {})
        self.colors = meta_data.get(META_DATA_KEY.COLORS, {})
        self.color = layer_data.get(SAVE_KEY.COLOR, None)
        self.lock = layer_data.get(SAVE_KEY.LOCK, False)
        self.locks = meta_data.get(META_DATA_KEY.LOCKS, {})
        self._input_metadata = meta_data
        self.spec_list = []
        self._nodes_path_as_key = {}
        self._nodes_node_as_key = {}
        self._construct_node_specs(layer_data)
        self.refresh()

    def get_cwd(self):
        path = getattr(self, SAVE_KEY.CWD)
        if path:
            return path
        path = self.real_path
        if path:
            return os.path.dirname(path)
        return os.getcwd()

    def _construct_node_specs(self, layer_data):
        self.spec_list = []
        self._nodes_path_as_key = {}
        self._nodes_node_as_key = {}
        nodes = order_nodes_dict(layer_data.get(SAVE_KEY.NODES, {}))
        for node_path, node_data in nodes.items():
            parent_path = nxt_path.get_parent_path(node_path)
            root_name = nxt_path.node_name_from_node_path(node_path)
            node_data['name'] = root_name
            node_data[nxt_node.INTERNAL_ATTRS.SOURCE_LAYER] = self.real_path
            root_spec_node = nxt_node.create_spec_node(node_data, self,
                                                       parent_path=parent_path)
            root_parent_path = getattr(root_spec_node,
                                       nxt_node.INTERNAL_ATTRS.PARENT_PATH)
            root_node_path = nxt_path.join_node_paths(root_parent_path,
                                                      root_name)
            self.spec_list += [root_spec_node]
            self._nodes_path_as_key[root_node_path] = root_spec_node
            self._nodes_node_as_key[root_spec_node] = root_node_path
            self.clear_node_child_cache(root_node_path)

    def refresh(self):
        """Re-populates the node table, and sets nodes' source layer.
        """
        self._node_table = []
        for path, node in self._nodes_path_as_key.items():
            for attr in get_node_local_attr_names(path, [self]):
                source_attr = attr + nxt_node.META_ATTRS.SOURCE
                setattr(node, source_attr, (self.real_path, path))
            node_ns = nxt_path.str_path_to_node_namespace(path)
            self._node_table += [[node_ns, node]]
            self.clear_node_child_cache(path)
        self.sort_node_table()

    def sort_node_table(self):
        self._node_table = sort_multidimensional_list(self._node_table,
                                                      sort_by_idx=0)

    def layer_idx(self):
        return self._layer_idx

    def lookup(self, node_path):
        return self._nodes_path_as_key.get(node_path)

    def node_exists(self, node_path):
        return bool(self.lookup(node_path))

    def get_exec_in(self, node_path):
        node = self.lookup(node_path)
        parent_path = getattr(node, nxt_node.INTERNAL_ATTRS.PARENT_PATH)
        if node and parent_path == nxt_path.WORLD:
            try:
                return getattr(node, nxt_node.INTERNAL_ATTRS.EXECUTE_IN)
            except AttributeError:
                pass
        return None

    def get_node_path(self, node):
        return self._nodes_node_as_key.get(node)

    def ancestors(self, node_path, return_type=LayerReturnTypes.Node,
                  include_implied=False):
        if include_implied and return_type != LayerReturnTypes.Path:
            raise TypeError('When including implied nodes, {} is an '
                            'unsupported return type'.format(return_type))
        if include_implied:
            return nxt_path.all_ancestor_paths(node_path)
        ancestor_nodes = []
        ancestor_paths = []
        node_table = []
        name_dict = {}
        if sys.version_info[0] == 2:
            node_path_is_str = isinstance(node_path, basestring)
        else:
            node_path_is_str = isinstance(node_path, str)
        if not node_path_is_str:
            node_path = self.get_node_path(node_path)
        node = self.lookup(node_path)
        parent_path = getattr(node, nxt_node.INTERNAL_ATTRS.PARENT_PATH)
        if node:
            while parent_path != nxt_path.WORLD:
                _node = self.lookup(parent_path)
                if _node is None:
                    parent_path = nxt_path.get_parent_path(parent_path)
                    continue
                if return_type == LayerReturnTypes.Boolean:
                    return True
                elif return_type == LayerReturnTypes.Node:
                    ancestor_nodes += [_node]
                elif return_type == LayerReturnTypes.Path:
                    ancestor_paths += [parent_path]
                elif return_type == LayerReturnTypes.NodeTable:
                    node_table += [[parent_path, _node]]
                elif return_type == LayerReturnTypes.NameDict:
                    key = getattr(node, nxt_node.INTERNAL_ATTRS.NAME)
                    name_dict[key] = _node
                node = _node
                ppath_attr = nxt_node.INTERNAL_ATTRS.PARENT_PATH
                parent_path = getattr(node, ppath_attr)

        if return_type == LayerReturnTypes.Node:
            return ancestor_nodes
        elif return_type == LayerReturnTypes.Path:
            return ancestor_paths
        elif return_type == LayerReturnTypes.NodeTable:
            return node_table
        elif return_type == LayerReturnTypes.NameDict:
            return name_dict
        elif return_type == LayerReturnTypes.Boolean:
            return False

    def children(self, node_path=nxt_path.WORLD,
                 return_type=LayerReturnTypes.Node, ordered=False,
                 include_implied=False):
        if include_implied and return_type != LayerReturnTypes.Path:
            raise ValueError('When including implied, can only return path. '
                             'Nothing else exists for implicit nodes.')
        children_nodes = []
        children_paths = []
        node_table = []
        name_dict = {}
        if not node_path:
            if return_type in (LayerReturnTypes.Node, LayerReturnTypes.Path,
                               LayerReturnTypes.NodeTable):
                return []
            elif return_type is LayerReturnTypes.NameDict:
                return {}
            elif return_type is LayerReturnTypes.Boolean:
                return False
            else:
                logger.error('Invalid return type provided')
                return None
        child_order = []
        implied_children = []
        # Look up real children cache
        children_cache = self._cached_children.get(node_path)
        if node_path == nxt_path.WORLD:
            children_cache = None
        if children_cache is not None:
            children_nodes = children_cache[LayerReturnTypes.Node][:]
            children_paths = children_cache[LayerReturnTypes.Path][:]
            node_table = children_cache[LayerReturnTypes.NodeTable][:]
            name_dict = copy.copy(children_cache[LayerReturnTypes.NameDict])
            cache_real = False
        else:
            self._cached_children[node_path] = {LayerReturnTypes.Node:
                                                children_nodes,
                                                LayerReturnTypes.Path:
                                                children_paths,
                                                LayerReturnTypes.NodeTable:
                                                node_table,
                                                LayerReturnTypes.NameDict:
                                                name_dict}
            cache_real = True
        # Lookup implied cache
        cache_implied = False
        if include_implied:
            if node_path is nxt_path.WORLD:
                implied_c_cache = None
            else:
                implied_c_cache = self._cached_implied_children.get(node_path)
            if implied_c_cache is not None:
                implied_children = implied_c_cache[LayerReturnTypes.Path][:]
                if return_type is LayerReturnTypes.Boolean and not cache_real:
                    return bool(implied_children + children_paths)
            else:
                path_implied = {LayerReturnTypes.Path: implied_children}
                self._cached_implied_children[node_path] = path_implied
                cache_implied = True
        re_cache = cache_implied or cache_real
        if re_cache:
            for path, node in self._nodes_path_as_key.items():
                if cache_real:
                    parent_path = getattr(node,
                                          nxt_node.INTERNAL_ATTRS.PARENT_PATH)
                    if parent_path is node_path:
                        children_nodes += [node]
                        children_paths += [path]
                        node_table += [[path, node]]
                        key = getattr(node, nxt_node.INTERNAL_ATTRS.NAME)
                        name_dict[key] = node
                if not include_implied or not cache_implied:
                    continue
                # TODO: Caching the ancestors could save around 51ms per loop
                if nxt_path.is_ancestor(path, node_path):
                    trim_depth = nxt_path.get_path_depth(node_path) + 1
                    trimmed = nxt_path.trim_to_depth(path, trim_depth)
                    if trimmed not in implied_children:
                        implied_children += [trimmed]
        if cache_real:
            k = LayerReturnTypes.Path
            self._cached_children[node_path][k] = children_paths[:]
        if include_implied:
            for imp in implied_children:
                if imp not in children_paths:
                    children_paths += [imp]
        if return_type is LayerReturnTypes.Boolean:
            return bool(children_paths)
        if ordered:
            node = self.lookup(node_path)
            if include_implied and not node:
                child_order = []
            else:
                co_attr = nxt_node.INTERNAL_ATTRS.CHILD_ORDER
                child_order = getattr(node, co_attr)
        if child_order:
            ordered_child_nodes = []
            ordered_child_paths = []
            ordered_node_table = []
            for child_name in child_order:
                # return type NODE
                for n in children_nodes:
                    c_name = getattr(n, nxt_node.INTERNAL_ATTRS.NAME)
                    if c_name is child_name:
                        ordered_child_nodes += [n]
                # return type PATH
                for p in children_paths:
                    if nxt_path.node_name_from_node_path(p) == child_name:
                        ordered_child_paths += [p]
                # return type TABLE
                for item in node_table:
                    p, n = item
                    if nxt_path.node_name_from_node_path(p) == child_name:
                        ordered_node_table += [item]
            # return type NODE
            for n in children_nodes:
                if n not in ordered_child_nodes:
                    ordered_child_nodes += [n]
            children_nodes = ordered_child_nodes
            # return type PATH
            for p in children_paths:
                if p not in ordered_child_paths:
                    ordered_child_paths += [p]
            children_paths = ordered_child_paths
            # return type TABLE
            for item in node_table:
                p, n = item
                if p not in ordered_node_table:
                    ordered_node_table += [item]
            node_table = ordered_node_table

        if return_type is LayerReturnTypes.Node:
            return children_nodes
        elif return_type is LayerReturnTypes.Path:
            return children_paths
        elif return_type is LayerReturnTypes.NodeTable:
            return node_table
        elif return_type is LayerReturnTypes.NameDict:
            return name_dict
        elif return_type is LayerReturnTypes.Boolean:
            return False
        else:
            logger.error('Invalid return type provided')
            return None

    def get_cached_child_paths(self, parent_path):
        """Get the cached child paths for the given parent path. If no cache
        exists, None is returned.

        :param parent_path: Path to node that has children
        :type parent_path: str
        :return: List of node paths or None
        :rtype: list | None
        """
        return self._cached_children.get(parent_path,
                                         {}).get(LayerReturnTypes.Path)

    def descendants(self, node_path=nxt_path.WORLD,
                    return_type=LayerReturnTypes.Path, ordered=False,
                    include_implied=False):
        if return_type not in (LayerReturnTypes.Node, LayerReturnTypes.Path,
                               LayerReturnTypes.NodeTable,):
            raise TypeError('Unsupported return type {}'.format(return_type))
        if node_path == nxt_path.WORLD:
            if ordered:
                raise ValueError('Cannot get ordered descendants of the world')
            return self._world_descendants(return_type,
                                           include_implied=include_implied)
        if ordered:
            if return_type is not LayerReturnTypes.Path:
                raise ValueError('Ordered descedants can only return paths.')
            return self._ordered_descendants(node_path,
                                             include_implied=include_implied)
        descendants = self.children(node_path, return_type,
                                    include_implied=include_implied)
        more_descendants = descendants
        while more_descendants:
            _temp = []
            if return_type == LayerReturnTypes.Path:
                for d_pth in more_descendants:
                    _temp += self.children(d_pth, return_type,
                                           include_implied=include_implied)
            elif return_type == LayerReturnTypes.NodeTable:
                for d_pth, d in more_descendants:
                    _temp += self.children(d_pth, return_type,
                                           include_implied=include_implied)
            descendants += _temp
            more_descendants = _temp
        return descendants

    def _ordered_descendants(self, node_path, include_implied=False):
        desc = []
        for child_path in self.children(node_path, ordered=True,
                                        return_type=LayerReturnTypes.Path,
                                        include_implied=include_implied):
            desc += [child_path]
            desc += self._ordered_descendants(child_path,
                                              include_implied=include_implied)
        return desc

    def _world_descendants(self, return_type, include_implied=False):
        if include_implied and return_type != LayerReturnTypes.Path:
            raise ValueError('When including implied, can only return path. '
                             'Nothing else exists for implicit nodes.')
        if return_type == LayerReturnTypes.Path:
            paths = list(self._nodes_path_as_key.keys())
            if nxt_path.WORLD in paths:
                paths.remove(nxt_path.WORLD)
            if not include_implied:
                return paths
            implied_paths = set()
            for real_path in paths:
                ancest_paths = nxt_path.all_ancestor_paths(real_path)
                implied_paths = implied_paths.union(ancest_paths)
            if nxt_path.WORLD in implied_paths:
                implied_paths.remove(nxt_path.WORLD)
            return implied_paths.union(paths)
        if return_type == LayerReturnTypes.Node:
            nodes = list(self._nodes_node_as_key.keys())
            world_node = self._nodes_path_as_key.get(nxt_path.WORLD)
            if world_node:
                nodes.remove(world_node)
            return nodes
        if return_type == LayerReturnTypes.NodeTable:
            node_table = []
            for path, node in self._nodes_path_as_key.items():
                if path is nxt_path.WORLD:
                    continue
                node_table += [[path, node]]
            return node_table
        raise TypeError('Unsupported return type {}'.format(return_type))

    def add_child_to_child_cache(self, parent_path, child_path, child_node):
        if not parent_path or parent_path == nxt_path.WORLD:
            return
        children_cache = self._cached_children.get(parent_path)
        if children_cache is not None:
            children_nodes = children_cache[LayerReturnTypes.Node][:]
            children_paths = children_cache[LayerReturnTypes.Path][:]
            node_table = children_cache[LayerReturnTypes.NodeTable][:]
            name_dict = copy.copy(children_cache[LayerReturnTypes.NameDict])
        else:
            children_nodes = []
            children_paths = []
            node_table = []
            name_dict = {}

        children_nodes += [child_node]
        children_paths += [child_path]
        node_table += [[child_path, child_node]]
        key = getattr(child_node, nxt_node.INTERNAL_ATTRS.NAME)
        name_dict[key] = child_node

        cache = {LayerReturnTypes.Node: children_nodes,
                 LayerReturnTypes.Path: children_paths,
                 LayerReturnTypes.NodeTable: node_table,
                 LayerReturnTypes.NameDict: name_dict}

        self._cached_children[parent_path] = cache
        return

    def remove_child_from_child_cache(self, parent_path, child_path,
                                      child_node):
        if parent_path is nxt_path.WORLD:
            return
        children_cache = self._cached_children.get(parent_path)
        if children_cache is None:
            return

        children_nodes = children_cache[LayerReturnTypes.Node]
        children_paths = children_cache[LayerReturnTypes.Path]
        node_table = children_cache[LayerReturnTypes.NodeTable]
        name_dict = children_cache[LayerReturnTypes.NameDict]

        try:
            children_nodes.remove(child_node)
        except ValueError:
            pass
        try:
            children_paths.remove(child_path)
        except ValueError:
            pass
        try:
            node_table.remove([child_path, child_node])
        except ValueError:
            pass
        key = getattr(child_node, nxt_node.INTERNAL_ATTRS.NAME)
        try:
            name_dict.pop(key)
        except KeyError:
            pass

    def clear_node_child_cache(self, parent_path):
        try:
            self._cached_children.pop(parent_path)
        except KeyError:
            pass
        try:
            self._cached_implied_children.pop(parent_path)
        except KeyError:
            pass

    def get_exec_order(self, start_path):
        if start_path == nxt_path.WORLD:
            return []
        start_root_path = nxt_path.get_root_path(start_path)
        exec_order = []
        start_found = False
        root_exec_order = self.get_root_exec_order(start_root_path)
        for root_path in root_exec_order:
            if not start_found and root_path == start_path:
                start_found = True
            node = self.lookup(root_path)
            enabled = nxt_node.get_node_enabled(node)
            if enabled is None:
                enabled = True
            if start_found and not enabled:
                continue
            if start_found:
                exec_order += [root_path]
            path_rt_type = LayerReturnTypes.Path
            disabled_desc = []
            for desc_path in self.descendants(node_path=root_path,
                                              ordered=True,
                                              return_type=path_rt_type):
                if not start_found and desc_path == start_path:
                    start_found = True
                desc_node = self.lookup(desc_path)
                enabled = nxt_node.get_node_enabled(desc_node)
                if enabled is None:
                    enabled = True
                if enabled:
                    for potential_ancestor in disabled_desc:
                        if nxt_path.is_ancestor(desc_path, potential_ancestor):
                            enabled = False
                            break
                if not enabled:
                    disabled_desc += [desc_path]
                    continue
                if start_found:
                    exec_order += [desc_path]
        return exec_order

    def get_root_exec_order(self, start_root_path):
        exec_order = [start_root_path]
        prev_root_path = start_root_path
        while True:
            for root in self.children():
                try:
                    exec_in = getattr(root, nxt_node.INTERNAL_ATTRS.EXECUTE_IN)
                    match = exec_in == prev_root_path
                except AttributeError:
                    continue
                if not match:
                    continue
                prev_root_path = self.get_node_path(root)
                exec_order += [prev_root_path]
                break  # for loop
            else:
                break  # while loop
        return exec_order

    def get_muted(self, local=False):
        muted = getattr(self, SAVE_KEY.MUTE, False)
        if not local:
            ref_data = get_comped_layer_overs(self, SAVE_KEY.COMP_ORVERRIDES)
            r_dat = ref_data.get(self.filepath, {})
            muted = r_dat.get(SAVE_KEY.MUTE, muted)
        return muted

    def set_muted(self, state):
        self.mute = state

    def set_mute_over(self, layer_path, state):
        ref_data = get_comped_layer_overs(self, SAVE_KEY.COMP_ORVERRIDES)
        r_dat = ref_data.get(layer_path, {})
        r_dat[SAVE_KEY.MUTE] = state
        ref_data[layer_path] = r_dat
        self.comp_overrides = ref_data

    def get_soloed(self, local=False):
        soloed = getattr(self, SAVE_KEY.SOLO, False)
        if not local:
            ref_data = get_comped_layer_overs(self, SAVE_KEY.COMP_ORVERRIDES)
            r_dat = ref_data.get(self.filepath, {})
            soloed = r_dat.get(SAVE_KEY.SOLO, soloed)
        return soloed

    def set_soloed(self, state):
        self.solo = state

    def set_solo_over(self, layer_path, state):
        ref_data = get_comped_layer_overs(self, SAVE_KEY.COMP_ORVERRIDES)
        r_dat = ref_data.get(layer_path, {})
        r_dat[SAVE_KEY.SOLO] = state
        ref_data[layer_path] = r_dat
        self.comp_overrides = ref_data

    def add_reference(self, layer_path=None, layer=None, insert_idx=None):
        """Given a layer path, a layer object, or both, add a reference to
        specified layer on this layer, optionally at a specified insert index.

        :param layer_path: Path to use as reference path, defaults to None, \
        if None is given, defaults to given `layer.real_path`
        :type layer_path: str, optional
        :param layer: layer to add as reference layer in this layer
        :type layer: SpecLayer
        :param insert_idx: index to insert given layer at, defaults to None, \
        if None is given, defaults to lowest reference layer.
        :type insert_idx: int, optional
        :return: index given layer was inserted at.
        :rtype: int
        :raises ValueError: If not provided with a layer path or layer.
        """
        if not (layer_path or layer):
            raise ValueError("Must specify layer path or layer to reference.")
        if insert_idx is None:
            insert_idx = len(self.sub_layer_paths)
        if layer_path is None:
            layer_path = layer.real_path
        self.sub_layer_paths.insert(insert_idx, layer_path)
        sub_layer_data = {SAVE_KEY.FILEPATH: layer_path}
        if layer is not None:
            sub_layer_data['layer'] = layer
        self.sub_layers.insert(insert_idx, sub_layer_data)

    def get_references(self):
        refs = []
        for ref in self.sub_layers:
            refs += [ref[SAVE_KEY.FILEPATH]]
        return refs

    def _get_meta_attr(self, meta_attr, over_dict_attr=None, local=False,
                       fallback_to_local=True):
        """Private method for getting meta attr opinions, comped or local,
        from the layer class.
        :param meta_attr: Class attr to get the opinion of (ie color or alias)
        :param over_dict_attr: Class attr name which corresponds to a dict of
        overrides for the given meta_attr on lower layers
        :param local: If Ture the local, non-comped, opinion is returned
        :param fallback_to_local: If True and there is no override the local
        opinion is returned.
        :raises: ValueError
        :return: opinion of meta_attr
        """
        val = getattr(self, meta_attr, None)
        if not local:
            both_or_none = (over_dict_attr, not local)
            if any(both_or_none) and not all(both_or_none):
                raise ValueError('Must provide an over_dict_attr if local is '
                                 'False!')
            overs = get_comped_layer_overs(self, over_dict_attr)
            if fallback_to_local:
                fallback = val
            else:
                fallback = None
            val = overs.get(self.filepath, fallback)
        return val

    def get_alias(self, local=False, fallback_to_local=True):
        """Get the layer's alias (nice name). By default the local alias is
        returned, if local is False the parent layer's opinion of the alias is
        returned.
        :param local: If True the local opinion is returned, if False the
        strongest override opinion is returned.
        :type local: bool
        :param fallback_to_local: if True and local is False and there is no
        override opinion the local alias is returned. If false and no
        override opinion is found None is returned.
        :return: string of layer alias or None
        """
        return self._get_meta_attr(SAVE_KEY.ALIAS, META_DATA_KEY.ALIASES,
                                   local=local,
                                   fallback_to_local=fallback_to_local)

    def set_alias(self, alias):
        self.alias = alias

    def set_alias_over(self, alias):
        """Set a alias override on the root parent of a layer.
        :param alias: new alias string
        """
        layer_path = self.filepath
        root = get_root_parent(self)
        if not root:
            logger.error('Can not set alias override on top layer!')
        if alias is None:
            if layer_path in list(root.aliases.keys()):
                root.aliases.pop(layer_path)
        else:
            root.aliases[layer_path] = alias

    def get_color(self, local=True, fallback_to_local=True):
        """Get the layer's color. By default the local color is
        retuned, if local is False the parent layer's opinion of the alias is
        returned.
        :param local: bool
        :param fallback_to_local: if True and local is False and there is no
        override opinion the local color is returned. If false and no
        override opinion is found None is returned.
        :return: string of layer alias
        """
        return self._get_meta_attr(SAVE_KEY.COLOR, META_DATA_KEY.COLORS,
                                   local=local,
                                   fallback_to_local=fallback_to_local)

    def set_color_over(self, color):
        """Set a color override on the root parent of a layer.
        :param color: hex color
        """
        layer_path = self.filepath
        root = get_root_parent(self)
        if not root:
            logger.error('Can not set colors override on top layer!')
        if color is None:
            if layer_path in list(root.colors.keys()):
                root.colors.pop(layer_path)
        else:
            root.colors[layer_path] = color

    def get_locked(self, local=False, fallback_to_local=True):
        """Get the locked state of the layer. By
        default the local opinion is returned, if local is False the parent
        layer's opinion of the alias is returned.
        :param local: If True the local opinion is returned, if False the
        strongest override opinion is returned.
        :type local: bool
        :param fallback_to_local: if True and local is False and there is no
        override opinion the local alias is returned. If false and no
        override opinion is found None is returned.
        :return: bool of layer's lock state
        """
        locked = self._get_meta_attr(SAVE_KEY.LOCK, META_DATA_KEY.LOCKS,
                                     local=local,
                                     fallback_to_local=fallback_to_local)
        return locked or False

    def set_locked(self, lock):
        self.lock = lock

    def set_locked_over(self, lock):
        """Set a lock override on the root parent of a layer.
        :param lock: bool
        """
        layer_path = self.filepath
        root = get_root_parent(self)
        if not root:
            logger.error('Can not set colors override on top layer!')
        if lock is None:
            if layer_path in list(root.colors.keys()):
                root.locks.pop(layer_path)
        else:
            root.locks[layer_path] = lock

    def save(self, filepath=None):
        """Save this layer, optionally to new, given filepath.

        :param filepath: file path to save to, defaults to None
        :type filepath: str, optional
        """
        filepath = filepath or self.real_path
        # Update graph name
        graph_name = os.path.splitext(os.path.basename(filepath))[0]
        if self.get_alias(local=True) == UNTITLED:
            self.set_alias(graph_name)
        save_data = self.get_save_data()
        try:
            json.dumps(save_data, indent=4, sort_keys=False)
        except TypeError:
            logger.error("Failed to save file!")
            # TODO: Should raise here, but its out of scope for what I'm
            #  working on right now.
            return
        logger.info("Save Data Generated")
        nxt_io.save_file_data(save_data=save_data, filepath=filepath)
        if filepath != self.real_path:
            self.real_path = filepath
            self.propegate_real_path()
        return save_data

    def get_meta_data(self):
        positions = OrderedDict(sorted(self.positions.items(),
                                       key=lambda x: x[0]))
        collapsed = OrderedDict(sorted(self.collapse.items(),
                                       key=lambda x: x[0]))
        aliases = OrderedDict(sorted(self.aliases.items(), key=lambda x: x[0]))
        colors = OrderedDict(sorted(self.colors.items(), key=lambda x: x[0]))
        locks = OrderedDict(sorted(self.locks.items(), key=lambda x: x[0]))
        meta_data = OrderedDict(self._input_metadata)
        if aliases:
            meta_data[META_DATA_KEY.ALIASES] = aliases
        if colors:
            meta_data[META_DATA_KEY.COLORS] = colors
        if locks:
            meta_data[META_DATA_KEY.LOCKS] = locks
        if positions:
            meta_data[META_DATA_KEY.POSITIONS] = positions
        if collapsed:
            meta_data[META_DATA_KEY.COLLAPSE] = collapsed
        return meta_data

    def get_comp_overrides(self):
        return OrderedDict(sorted(self.comp_overrides.items(),
                                  key=lambda x: x[0]))

    def get_save_data(self):
        """Returns save data for this layer.
        """
        save_dict = {
            SAVE_KEY.VERSION: GRAPH_VERSION.VERSION_STR,
            SAVE_KEY.ALIAS: self.get_alias(local=True),
            SAVE_KEY.MUTE: self.get_muted(local=True),
            SAVE_KEY.SOLO: self.get_soloed(local=True),
            SAVE_KEY.REFERENCES: self.get_references(),
            SAVE_KEY.COMP_ORVERRIDES: self.get_comp_overrides(),
            SAVE_KEY.COLOR: self.color,
            SAVE_KEY.META_DATA: self.get_meta_data(),
            SAVE_KEY.NODES: self.get_nodes_save_data()}
        rm = []
        for k, v in save_dict.items():
            if v in ((), [], {}):
                rm += [k]
        for k in rm:
            save_dict.pop(k)
        return self.order_save_dict(save_dict)

    def order_save_dict(self, save_dict):
        """Given a save dictionary, returns an ordered dictionary where the
        save keys are ordered correctly. Leftover keys are put at the end.
        :param data: dictionary to order
        :type data: dict
        :return: ordered version of given dictionary
        :rtype: OrderedDcit
        """
        # order keys
        keys_order = (SAVE_KEY.VERSION, SAVE_KEY.ALIAS, SAVE_KEY.COLOR,
                      SAVE_KEY.MUTE, SAVE_KEY.SOLO, SAVE_KEY.REFERENCES,
                      SAVE_KEY.COMP_ORVERRIDES, SAVE_KEY.META_DATA,
                      SAVE_KEY.NODES, SAVE_KEY.REAL_PATH)

        result = OrderedDict()
        data_keys = list(save_dict.keys())
        for key in keys_order:
            if key in data_keys:
                result[key] = save_dict[key]
                data_keys.remove(key)
        # leftovers
        for key in data_keys:
            result[key] = save_dict[key]
        return result

    def get_nodes_save_data(self):
        """Gets an ordered dict of all of the spec nodes
        :return: OrderedDict
        """
        nodes_dict = OrderedDict()
        for spec_node in self.spec_list:
            node_data = nxt_node.get_node_as_dict(spec_node)
            node_path = self.get_node_path(spec_node)
            nodes_dict[node_path] = node_data
        return order_nodes_dict(nodes_dict)

    def propegate_real_path(self):
        """Update nodes source layer to current real path
        """
        for node in self.descendants(return_type=LayerReturnTypes.Node):
            setattr(node, nxt_node.INTERNAL_ATTRS.SOURCE_LAYER, self.real_path)


class CompLayer(SpecLayer):
    @classmethod
    def load_from_layer_data(cls, layer_data):
        raise NotImplementedError

    def __init__(self):
        super(CompLayer, self).__init__({})
        self._name = 'CompositeLayer'
        self._layer_range = (0, 0)
        self._dirty_map = {}
        self._sublayer_node_tables = []
        self.runtime = False
        self.running = False
        self._console = Console(_globals={}, _locals={}, node_path='STAGE')
        self.cache_layer = CacheLayer()
        # Quick and dirty hack to be used by the UI
        self.failure = False

    def layer_idx(self):
        return self._layer_range[0]

    def get_node_dirties(self, node_path):
        """Get all nodes that depend on the given node path. The given node
        path will be the 0th item in the return list.
        :param node_path: String of node path
        :return: List of string node paths
        """
        concerns = self._dirty_map.get(node_path, [])[:]
        new_concerns = [node_path] + concerns
        while new_concerns:
            _tmp = []
            for concern in new_concerns:
                found = self._dirty_map.get(concern, [])
                for item in found:
                    if item not in concerns + _tmp:
                        _tmp += [item]
            new_concerns = _tmp
            concerns += _tmp
        return concerns


class CacheLayer(SpecLayer):
    def __init__(self):
        super(CacheLayer, self).__init__({})
        self.nodes = {}
        self.node_time_map = {}
        self.node_times = []

    def was_during_node_exec(self, time_float):
        """Checks if given time float is within any executed node's exec
        timeframe.
        :param time_float: float of time
        :return: bool
        """
        for enter_time, exit_time in reversed(self.node_times):
            if time_float >= enter_time and exit_time is None:
                return True
            if enter_time <= time_float <= exit_time:
                return True
        return False

    def set_node_enter_time(self, node_path, t=0.):
        if not t:
            t = time.time()
        node_runs = self.node_time_map.get(node_path, [])
        time_entry = [[t, None]]
        node_runs += time_entry
        self.node_times += time_entry
        self.node_time_map[node_path] = node_runs

    def set_node_exit_time(self, node_path, t=0.):
        if not t:
            t = time.time()
        try:
            node_runs = self.node_time_map[node_path]
        except IndexError:
            raise IndexError('No start time set for {}'.format(node_path))
        for run_time in node_runs:
            if run_time[1] is None:
                run_time.pop(1)
                run_time += [t]
                return
        raise ValueError('No exec tuples for {} have a missing end '
                         'time!'.format(node_path))

    def get_node_run_time(self, node_path, idx=-1):
        try:
            enter_time, exit_time = self.node_time_map[node_path][idx]
        except (KeyError, IndexError):
            raise ValueError('No time for {} at index {}'.format(node_path,
                                                                 idx))
        return exit_time - enter_time

    def lookup(self, node_path):
        return self.nodes.get(node_path)

    def add_node(self, node_path, node):
        cached_code = getattr(node, nxt_node.INTERNAL_ATTRS.CACHED_CODE, '\n')
        spl_code = cached_code.split('\n')
        setattr(node, nxt_node.INTERNAL_ATTRS.COMPUTE, spl_code)
        self.nodes[node_path] = node

    def save(self, filepath):
        save_data = self.get_save_data()
        nxt_io.save_file_data(save_data=save_data, filepath=filepath)
        return save_data

    def get_save_data(self):
        save_data = {
            'version': GRAPH_VERSION.VERSION_STR,
            'nodes': {}
        }
        for path, node in self.nodes.items():
            save_data['nodes'][path] = nxt_node.get_node_as_dict(node)
        return save_data

    @classmethod
    def load_from_layer_data(cls, layer_data):
        new_layer = cls()
        result_nodes = {}
        name_attr = nxt_node.INTERNAL_ATTRS.NAME
        name_key = nxt_node.INTERNAL_ATTRS.as_save_key(name_attr)
        for node_path, node_data in layer_data.get('nodes', {}).items():
            parent_path = nxt_path.get_parent_path(node_path)
            node_name = nxt_path.node_name_from_node_path(node_path)
            node_data[name_key] = node_name
            spec_node = nxt_node.create_spec_node(node_data, new_layer,
                                                  parent_path)
            code_lines = getattr(spec_node, nxt_node.INTERNAL_ATTRS.COMPUTE)
            code = '\n'.join(code_lines)
            setattr(spec_node, nxt_node.INTERNAL_ATTRS.CACHED_CODE, code)
            result_nodes[node_path] = spec_node
        new_layer.nodes = result_nodes
        return new_layer


def get_active_layers(layers):
    """Given a list of layers, return those that should contribute to comp.

    If there are soloed layers, only the soloed layers that are not muted
    will be returned. If there are not soloed layers, only the non-muted
    layers will be returned.

    :param layers: list of layers to filter
    :type layers: list
    :return: list of layers that should contribute to a comp.
    :rtype: list
    """
    soloed_layers = get_soloed_layers(layers)
    potential_layers = soloed_layers or layers
    result_layers = []
    for layer in potential_layers:
        if layer.get_muted():
            continue
        result_layers += [layer]
    return result_layers


def get_soloed_layers(layers):
    """Given a list of layers, return only those that are soloed

    :param layers: list of layers to filter
    :type layers: list
    :return: list of layers that are soloed.
    :rtype: list
    """
    return [layer for layer in layers if layer.get_soloed()]


def get_muted_layers(layers):
    """Given a list of layers, return only those that are muted

    :param layers: list of layers to filter
    :type layers: list
    :return: list of layers that are muted.
    :rtype: list
    """
    return [layer for layer in layers if layer.get_muted()]


def get_node_local_attr_names(node_path, layers):
    """Get all attribute names that are local to the given node path within
    the given layers.

    NOTE that layers are filtered to active layers.

    :param node_path: node path to find local attributes of
    :type node_path: str
    :param layers: list of layers to look for attributes of given node within
    :type layers: list
    :return: list of local attributes found in given layers
    :rtype: list
    """
    local_attrs = []
    # TODO should we filter for active layers or expect callers to?
    active_layers = get_active_layers(layers)
    for layer in active_layers:
        node = layer.lookup(node_path)
        if not node:
            continue
        for attr in node.__dict__.keys():
            if attr in local_attrs:
                continue
            if attr in nxt_node.INTERNAL_ATTRS.PROTECTED:
                continue
            if attr.endswith(nxt_node.META_ATTRS._suffix):
                continue
            local_attrs += [attr]
    return local_attrs


def get_comped_layer_overs(layer, over_dict_attr):
    over_dict = {}
    over_dict.update(getattr(layer, over_dict_attr, {}))
    parent = layer.parent_layer
    while parent:
        over_dict.update(getattr(parent, over_dict_attr, {}))
        parent = parent.parent_layer
    return over_dict


def get_root_parent(layer):
    parent = layer.parent_layer
    root = None
    while parent:
        root = parent
        parent = parent.parent_layer
    return root


def order_nodes_dict(node_dict):
    """Given a dictionary mapping node paths to node data, return an ordered
    dict where the keys are sorted.

    :param node_dict: dictionary mapping node paths to node data
    :type node_dict: dict
    :return: ordered dctionary with keys sosrted
    :rtype: OrderedDict
    """
    unsorted_paths = list(node_dict.keys())
    sorted_paths = sorted(unsorted_paths)
    ordered_nodes = OrderedDict()
    for path in sorted_paths:
        ordered_nodes[path] = node_dict[path]
    return ordered_nodes


def sort_multidimensional_list(multi_list, sort_by_idx):
    """Takes a multi-dimensional list and sorts it by the length of a sub-list
    item. The item who's length we sort by is
    determined by the idx parameter.
    If your input is:
    ```
    sort_multidimensional_list(
    [
        [[1,2,3], ['a', 'b']],
        [[4,5,6], ['a']]
    ], idx = 1)
    ```
    Your output will be:
    ```
    [
        [[1,2,3], ['a']],
        [[4,5,6], ['a', 'b']]
    ]
    ```
    :param multi_list: multi-dimensional list
    :param sort_by_idx: Index of sub list item to sort by
    :return: list
    """
    if not multi_list:
        return multi_list
    list_len = len(multi_list)
    i = 0
    while i <= list_len:
        ii = 0
        ii_high_end = list_len - i - 1
        item_len = len(multi_list[0][sort_by_idx])
        while ii < ii_high_end:
            ii_plus_one = ii + 1
            next_item_len = len(multi_list[ii_plus_one][sort_by_idx])
            if item_len > next_item_len:
                _temp = multi_list[ii]
                multi_list[ii] = multi_list[ii_plus_one]
                multi_list[ii_plus_one] = _temp
            else:
                item_len = next_item_len
            ii = ii_plus_one
        i += 1
    return multi_list



# Builtin
import unittest

# Internal
from nxt.session import Session
from nxt.test import get_test_file_path
from nxt import DATA_STATE

class TestResolve(unittest.TestCase):
    def test_multiline_code_resolve(self):
        """When an attr has "real" newline and is subsituted into code,
        resolve that into multiple code lines
        """
        test_graph = get_test_file_path("ResolveNewlines.nxt")
        stage = Session().load_file(test_graph)
        comp_layer = stage.build_stage()
        node=comp_layer.lookup("/test_node")
        # Unresolved has 1 line
        raw_lines = stage.get_node_code_lines(node,
                                              layer=comp_layer,
                                              data_state=DATA_STATE.RAW)
        assert len(raw_lines) == 1
        resolved_lines = stage.get_node_code_lines(node,
                                                   layer=comp_layer,
                                                   data_state=DATA_STATE.RESOLVED)
        assert len(resolved_lines) == 2



# Builtin
import os
import json
import logging
import socket
import pickle

# Internal
from nxt.stage import Stage
from nxt import nxt_io
from nxt.runtime import GraphError

logger = logging.getLogger('nxt')

HOST = 'localhost'
CMD_PORT = 4435  # Command port number
COM_PORT = 4436  # Communication port number

HEADER_SIZE = 10  # Fixed header len
HANDSHAKE_TIMEOUT = 60  # Seconds to wait before giving up on handshake
# Global objects
global __nxt_model__
__nxt_model__ = None
MODEL_VAR = '__nxt_model__'
global __nxt_server__
__nxt_server__ = None
SERVER_VAR = '__nxt_server__'


class COM_TYPE(object):
    """Constants used as keys in the pickled message dicts sent over to the
    nxt socket server."""
    LOG = 'log'
    CACHE = 'cache'
    SHUTDOWN = 'shutdown'
    PING = 'ping'
    WAIT = 'wait'
    ERR = 'error'


def setup():
    """Setup a connection to the nxt socket server. Sets the global variable
    __nxt_server__ to the newly created socket instance.
    :return: socket instance
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.setblocking(1)
    s.connect((HOST, COM_PORT))
    logger.socket('Connected to: {}:{}'.format(HOST, COM_PORT))
    global __nxt_server__
    __nxt_server__ = s
    return s


def format_msg(msg, com_typ, msg_dict=None):
    """Formats the given msg into a dict matching the expected format used by
    nxt socket servers and clients.
    :param msg: Object to be formatted as a socket message
    :param com_typ: COM_TYPE constant
    :param msg_dict: Optionally if you are combining commands you can pass an
    existing msg_dict returned from a previous call to this function.
    :return: dict
    """
    if not msg_dict:
        msg_dict = {}
    msg_dict[com_typ] = msg
    return msg_dict


def get_nxt_model():
    """Gets the existing SocketClientModel instance in global scope, if none is
    found a new one is created and returned. It is stored in the global scope
    to allow maximum context compatibility.
    :return: SocketClientModel instance
    """
    global __nxt_model__
    if __nxt_model__ is None:
        m = SocketClientModel(None)
        __nxt_model__ = m
    return __nxt_model__


class SocketClientModel(object):
    """Simple nxt model used by remote socket clients"""
    def __init__(self, com_server):
        self.filepath = None
        self.exec_order = None
        self.param_filepath = None
        self.file_data = {}
        self.stage = None
        self.comp_layer = None
        self.runtime_layer = None
        self.cache_filepath = None
        self.server = com_server

    def load(self, filepath):
        """Loads the given file path and builds a comp layer ready to be
        executed.
        :param filepath: string of nxt save filepath.
        :return: None
        """
        with IPCWait(self):
            logger.socket('Loading file data from {}...'.format(filepath))
            self.filepath = filepath
            self.file_data = nxt_io.load_file_data(self.filepath)
            logger.socket('Building stage...')
            self.stage = Stage(self.file_data)
            self._rebuild_comp()

    def _run(self, exec_order=()):
        """Private method for executing nodes, not meant to be called outside
        of a IPCWait context manager.
        :param exec_order: Optional list of node paths to execute.
        :raises GraphError
        :return: True if no errors were detected.
        """
        if exec_order:
            self.exec_order = exec_order
        if not self.stage:
            logger.error('No stage built, cannot run!')
            return False
        parameters = {}
        if os.path.isfile(self.param_filepath or ''):
            with open(self.param_filepath, 'r') as fp:
                parameters = json.load(fp)
        if not self.runtime_layer:
            self._rebuild_comp()
        self.runtime_layer = self.stage.execute_nodes(self.exec_order,
                                                      self.runtime_layer,
                                                      parameters=parameters)
        return True

    def run(self, exec_order=()):
        """General purpose run function meant to be called over a socket
        command port. If a GraphError is raised the internal nxt ERR message
        is sent over the socket to the server.
        :param exec_order: Optional list of node paths to execute.
        :return: None
        """
        with IPCWait(self):
            try:
                self._run(exec_order)
            except GraphError:
                send_to_server(format_msg(None, COM_TYPE.ERR))

    def refresh(self):
        """Refresh (rebuild) stage and comp objects.
        :return: None
        """
        with IPCWait(self):
            if not self.filepath:
                logger.error('No filepath, cannot refresh!')
                return
            self.file_data = nxt_io.load_file_data(self.filepath)
            self.stage = Stage(self.file_data)
            self._rebuild_comp()

    def _rebuild_comp(self):
        """Private method for rebuilding comp layer. Not meant to be called
        oustide of a IPCWait context manger.
        :return: None
        """
        with IPCWait(self):
            if not self.stage:
                logger.error('No data loaded, cannot re-build!')
                return

            logger.compinfo('Rebuilding comp...')
            self.comp_layer = self.stage.build_stage()
            logger.compinfo('Rebuilding runtime layer...')
            self.runtime_layer = self.stage.setup_runtime_layer(self.comp_layer)

    def close(self, notify_server=False):
        """Attempts to gracefully close connection to nxt socket server."""
        # with IPCWait(self):
        for log_handler in logger.handlers:
            if isinstance(log_handler, NXTSocketLogHandler):
                logger.removeHandler(log_handler)
                logger.debug('Removed nxt log handler')
                break
        if notify_server:
            send_to_server(format_msg(None, COM_TYPE.SHUTDOWN))
            logger.socket('Telling nxt server to close connection!')
        global __nxt_server__
        if __nxt_server__:
            __nxt_server__.close()
        __nxt_server__ = None
        self.server = __nxt_server__

    def open(self):
        """Opens connection to nxt socket server. Installs
        NXTSocketLogHandler logging handler if none is installed.
        :return: None
        """
        logger.socket('Connecting to listener...')
        self.server = setup()
        global __nxt_server__
        __nxt_server__ = self.server
        setup_socket_log = True
        for log_handler in logger.handlers:
            if isinstance(log_handler, NXTSocketLogHandler):
                setup_socket_log = False
                break
        if setup_socket_log:
            logger.debug('Making new log handler!')
            logger.addHandler(NXTSocketLogHandler())
        logger.socket('Successfully connected to listener!')

    def ping(self):
        """Debug function for testing 2 way communication from the nxt socket
        server to this client.
        :return:
        """
        logger.socket('Pinging...')
        with IPCWait(self):
            send_to_server(format_msg('', COM_TYPE.PING))
        logger.socket('Pinged!')

    def get_cache(self):
        """Method for the nxt socket server to request cache data from this
        client model.
        :return: Save data dict
        """
        with IPCWait(self):
            save_data = self.runtime_layer.cache_layer.save(self.cache_filepath)
            send_to_server(format_msg(save_data, COM_TYPE.CACHE))
        return save_data


def send_log(record):
    """Simple method for sending log records to an nxt socket server.
    :param record: logging record
    :return: None
    """
    links = getattr(record, 'links', None)
    data_dict = format_msg([record.levelno, record.message, links],
                           COM_TYPE.LOG)
    send_to_server(data_dict)


def send_to_server(data_dict):
    """Pickles the given data, assembles fixed length header, casts to bytes
    and sends to the nxt socket server.
    NOTE: This function catches ALL exceptions raised by the socket instance!
    :param data_dict: Dict formatted by format_msg
    :return: None
    """
    # Don't use logging in here as it can cause in infinite loop. The logging
    # handler calls this function.
    global __nxt_server__
    if not __nxt_server__:
        logger.error("No server! Try toggling the command port.")
        return
    msg = pickle.dumps(data_dict)
    header = bytes("{:<{}}".format(len(msg), HEADER_SIZE))
    data = header + msg
    # logger.debug('Size of data: {}'.format(sys.getsizeof(data)))
    try:
        __nxt_server__.sendall(data)
    except Exception as e:
        if e.errno == 10053:  # Stale socket
            __nxt_server__ = None
        else:
            logger.exception('Failed to send data to socket server!')


class IPCWait(object):
    """Simple context manager for telling the server to wait while a command
    is being processed by the client interpreter.
    """
    def __init__(self, model):
        self.model = model
        global __nxt_server__
        self.server = __nxt_server__

    def send_wait_msg(self, state):
        if not self.server:
            logger.error("No server, context manager doesn't know what to do!")
            return
        # logger.debug('Telling nxt to wait {}'.format(state))
        send_to_server(format_msg(state, COM_TYPE.WAIT))
        self.server.settimeout(HANDSHAKE_TIMEOUT)
        try:
            okay = self.server.recv(1024)
        except Exception as e:
            logger.exception('Failed to receive data from server! Try '
                             'toggling the command port.')
            return
        self.server.settimeout(None)
        # logger.debug('Server gave okay msg: {}'.format(okay))
        if not okay:
            logger.error('Failed to handshake!')
            send_to_server(format_msg(False, COM_TYPE.WAIT))
            return

    def __enter__(self):
        self.send_wait_msg(True)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.send_wait_msg(False)


class NXTSocketLogHandler(logging.Handler):
    """Simple logging handler for emitting log records to the nxt socket
    server.
    """
    def __init__(self):
        logging.Handler.__init__(self)

    def emit(self, record):
        send_log(record)



import setuptools
import json
import io

with io.open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

desc = ("A general purpose code compositor designed for rigging, "
        "scene assembly, and automation. (node execution tree)")

with open("nxt/version.json", 'r') as fh:
    version_data = json.load(fh)
api_v_data = version_data['API']
api_major = api_v_data['MAJOR']
api_minor = api_v_data['MINOR']
api_patch = api_v_data['PATCH']
api_version = '{}.{}.{}'.format(api_major, api_minor, api_patch)

setuptools.setup(
    name="nxt-core",
    version=api_version,
    author="The nxt contributors",
    author_email="dev@opennxt.dev",
    description=desc,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nxt-dev/nxt",
    packages=setuptools.find_packages(),
    python_requires='>=2.7, <3.11',
    entry_points={
        'console_scripts': [
            'nxt=nxt.cli:main',
        ],
    },
    package_data={
        # covers text nxt files
        "": ["*.nxt"],
        # Covers builtin, and full depth of test graphs/files
        "nxt": ["version.json",
                "builtin/*",
                "test/*",
                "test/spec_graphs/*",
                "test/legacy/*",
                "test/plugins/*",
                "test/plugins/fallbacks/*",
                "test/plugins/fallbacks/another/*",
                "test/plugins/fallbacks/base/*"
                ],
    }
)



# Builtin
import sys
if sys.version_info[0] == 2:
    from xmlrpclib import ServerProxy
else:
    from xmlrpc.client import ServerProxy

# Internal
from nxt.remote import get_running_server_address


class NxtClient(ServerProxy, object):
    def __init__(self, address=None):
        if address is None:
            address = 'http://{}'.format(get_running_server_address())
        self.address = address
        super(NxtClient, self).__init__(self.address, allow_none=True)

    def __repr__(self):
        return self.address



# Builtin
import sys
import os
import logging

# Internal Logging setup
if sys.version_info[0] == 2:
    import nxt_log
else:
    from . import nxt_log
nxt_log.initial_setup()

# Internal
if sys.version_info[0] == 2:
    import plugin_loader
    from constants import (DATA_STATE, NODE_ERRORS, UNTITLED, IGNORE,
                           GRID_SIZE, USER_PLUGIN_DIR, USER_DIR_ENV_VAR)
    from session import Session
    import nxt_path
    import nxt_layer
    from remote import contexts
else:
    from . import plugin_loader
    from .constants import (DATA_STATE, NODE_ERRORS, UNTITLED, IGNORE,
                            GRID_SIZE, USER_PLUGIN_DIR, USER_DIR_ENV_VAR)
    from .session import Session
    from . import nxt_path, nxt_layer
    from .remote import contexts

# Load plugins
plugin_loader.load_plugins()

logger = logging.getLogger('nxt')


def execute_graph(filepath, start=None, parameters=None, context=None):
    """Shortest code path to executing a graph from the nxt package.
    Creates a 1 off session and executes the graph within that session.
    Arguments are a direct copy of Session.execute_graph, see there for full
    details.

    :param filepath: Path to the graph file.
    :type filepath: str
    :param start: Path to the node to begin execution from.
    :type start: str
    :param parameters: Dict where key is attr path and value is new attr\
    value.
    :type parameters: dict
    :param context: Optional name of remote context to execute graph in, \
    if none is passed the graph is executed in this interpreter.
    :type context: str
    """
    if context and context not in contexts.iter_context_names():
        logger.info('Valid contexts are: '
                    '{}'.format(list(contexts.iter_context_names())))
        raise NameError('Unknown context: "{}"'.format(context))
    one_shot_session = Session()
    one_shot_session.execute_graph(filepath,
                                   start=start, parameters=parameters,
                                   context=context)


def create_context(custom_name, interpreter_exe=sys.executable,
                   context_graph=None, exe_script_args=()):
    """Generate a remote context given the path to an interpreter executable,
    if none is provided the current sys.executable is used. You must provide a
    custom name for your context. If no errors are raised, the context should
    be available without restarting NXT.

    :param interpreter_exe: String path to Python executable.  If another\
    executable is supplied a script to call with that executable is expected.
    :param custom_name: String of context name. The name users will refer to\
    the context by
    :param context_graph: Optional path to an existing context startup graph
    :param exe_script_args: If the exe arg is not a Python interpreter \
    additional args may need to passed. For example in Blender its possible \
    to call the blender.exe with the arg `--python`. After the args the path \
    to the nxt cli module is inserted into the subprocess call.
    :type exe_script_args: list | tuple
    :raises: IOError, NameError
    :return: string of the new context name
    """
    _user_plugin_dir = USER_PLUGIN_DIR.replace(os.sep, '/')
    if not custom_name:
        raise NameError('Must provide a valid name for your context!')
    interpreter_exe = interpreter_exe.replace(os.sep, '/')
    default_context_graph = 'custom_{name}_context.nxt'.format(name=custom_name)
    default_c_path = os.path.abspath(os.path.join(_user_plugin_dir,
                                                  default_context_graph))
    default_c_path = default_c_path.replace(os.sep, '/')
    template_path = os.path.join(os.path.dirname(__file__),
                                 'remote/remote_context_template.py')
    with open(template_path, 'r') as fp:
        py_str = fp.read()

    context_py_file = 'custom_{name}_context.py'.format(name=custom_name)
    context_py_filepath = os.path.join(_user_plugin_dir, context_py_file)
    if os.path.exists(context_py_filepath):
        raise IOError('The context name you selected is already taken. '
                      '"{}"'.format(custom_name))
    if context_graph is None:
        context_graph = default_c_path
        file_name = 'custom_{name}_context.nxt'.format(name=custom_name)
        dest_path = os.path.join(_user_plugin_dir, file_name).replace(os.sep,
                                                                     '/')
        exe_name = contexts.get_current_context_exe_name()
        starter_context = contexts.starter_contexts.get(exe_name)
        if starter_context is None:
            fallback_exe_name = os.path.basename(interpreter_exe.lower())
            fallback_exe_name, _ = os.path.splitext(fallback_exe_name)
            starter_context = contexts.starter_contexts.get(fallback_exe_name)
        if starter_context:
            layer = nxt_layer.SpecLayer.load_from_filepath(starter_context)
        else:
            logger.warning('No prebuilt context was found for "{exe_name}", '
                           'if your interpreter needs special setup/teardown '
                           'please open "{dest_path}" and edit as needed.'
                           ''.format(exe_name=exe_name, dest_path=dest_path))
            layer = nxt_layer.SpecLayer()
            layer.add_reference(layer_path=contexts._context_graph)
        layer.set_alias('custom_{name}_context'.format(name=custom_name))
        layer.save(filepath=dest_path)
    else:
        expanded_context_graph = nxt_path.full_file_expand(context_graph)
        if not os.path.isfile(expanded_context_graph):
            raise IOError('The context graph you provided does not exist! '
                          '"{}"'.format(context_graph))
        root = os.path.commonprefix((context_graph, context_py_filepath))
        if root:
            context_graph = os.path.relpath(context_graph, root)
        context_graph.replace(os.sep, '/')

    py_str = py_str.format(name=custom_name, interpreter_exe=interpreter_exe,
                           context_graph=context_graph, args=exe_script_args)
    with open(context_py_filepath, 'w+') as fp:
        fp.write(py_str)
    exec(py_str)
    return custom_name



# Builtin
import os
import unittest

# Internal
from nxt import stage, nxt_layer


class TestReferences(unittest.TestCase):
    def test_reference_by_path(self):
        test_dir = os.path.dirname(__file__)
        empty_path = os.path.join(test_dir, 'empty.nxt')
        pre_test = stage.Stage.load_from_filepath(empty_path).build_stage()
        # assert that empty is empty
        self.assertEqual(0, len(pre_test.descendants()))
        # Test adding reference
        empty_spec_layer = nxt_layer.SpecLayer.load_from_filepath(empty_path)
        empty_spec_layer.add_reference('ref_test.nxt')
        temporary_graph_path = os.path.join(test_dir, 'IWILLBEDELTED.nxt')
        empty_spec_layer.save(temporary_graph_path)
        # Rebuild stage and verify
        stage_with_ref = stage.Stage.load_from_filepath(temporary_graph_path)
        comp_layer_with_ref = stage_with_ref.build_stage()
        # Remove before asserting, to clean up even on failure.
        os.remove(temporary_graph_path)
        self.assertIsNotNone(comp_layer_with_ref.lookup('/i_am_here'))

    def test_reference_by_obj(self):
        test_dir = os.path.dirname(__file__)
        empty_path = os.path.join(test_dir, 'empty.nxt')
        pre_test = stage.Stage.load_from_filepath(empty_path).build_stage()
        # assert that empty is empty
        self.assertEqual(0, len(pre_test.descendants()))
        # Test adding reference
        empty_spec_layer = nxt_layer.SpecLayer.load_from_filepath(empty_path)
        ref_path = os.path.join(test_dir, 'ref_test.nxt')
        ref_test_spec_layer = nxt_layer.SpecLayer.load_from_filepath(ref_path)
        empty_spec_layer.add_reference(layer=ref_test_spec_layer)
        temporary_graph_path = os.path.join(test_dir, 'IWILLBEDELTED.nxt')
        empty_spec_layer.save(temporary_graph_path)
        # Rebuild stage and verify
        stage_with_ref = stage.Stage.load_from_filepath(temporary_graph_path)
        comp_layer_with_ref = stage_with_ref.build_stage()
        # Remove before asserting, to clean up even on failure.
        os.remove(temporary_graph_path)
        self.assertIsNotNone(comp_layer_with_ref.lookup('/i_am_here'))

